#include "rpc.h"
#include "relays.h"
#include "buttons.h"
#include "helper.h"

static void rpc_log(struct mg_rpc_request_info *ri, struct mg_str args) {
  LOG(LL_INFO, ("tag=%.*s src=%.*s method=%.*s args='%.*s'",
                ri->tag.len, ri->tag.p, ri->src.len, ri->src.p,
                ri->method.len, ri->method.p, args.len, args.p));
  // TODO(pim): log to MQTT
}


static void rpc_button_touch_handler(struct mg_rpc_request_info *ri, void *cb_arg, struct mg_rpc_frame_info *fi, struct mg_str args) {
  int idx;
  uint8_t gpio;

  rpc_log(ri, args);

  if (json_scanf(args.p, args.len, ri->args_fmt, &idx) != 1) {
    mg_rpc_send_errorf(ri, 400, "idx is required");
    ri = NULL;
    return;
  }

  if (ERR_GPIO_INVALID==(gpio=buttons_get_gpio_by_idx(idx))) {
    mg_rpc_send_errorf(ri, 400, "No button at idx=%d", idx);
    ri = NULL;
    return;
  }

  buttons_touch_by_idx(idx);
  mg_rpc_send_responsef(ri, "{idx: %d, gpio: %d}", idx, gpio);
  ri = NULL;

  (void) ri;
  (void) cb_arg;
  (void) fi;
  (void) args;
}

static void rpc_relay_get_handler(struct mg_rpc_request_info *ri, void *cb_arg, struct mg_rpc_frame_info *fi, struct mg_str args) {
  int idx;
  uint8_t gpio;
  bool state;

  rpc_log(ri, args);

  if (json_scanf(args.p, args.len, ri->args_fmt, &idx) != 1) {
    mg_rpc_send_errorf(ri, 400, "idx is required");
    ri = NULL;
    return;
  }

  if (ERR_GPIO_INVALID==(gpio=relays_get_gpio_by_idx(idx))) {
    mg_rpc_send_errorf(ri, 400, "No relay at idx=%d", idx);
    ri = NULL;
    return;
  }

  state=relays_get_by_idx(idx);
  mg_rpc_send_responsef(ri, "{idx: %d, gpio: %d, state: %d}", idx, gpio, state);
  ri = NULL;

  (void) cb_arg;
  (void) fi;
}

static void rpc_relay_set_handler(struct mg_rpc_request_info *ri, void *cb_arg, struct mg_rpc_frame_info *fi, struct mg_str args) {
  int idx;
  uint8_t gpio;
  int state;

  rpc_log(ri, args);

  if (json_scanf(args.p, args.len, ri->args_fmt, &idx, &state) != 2) {
    mg_rpc_send_errorf(ri, 400, "idx and state are required");
    ri = NULL;
    return;
  }
  if (state!=0 && state!=1) {
    mg_rpc_send_errorf(ri, 400, "state must be 0 or 1");
    ri = NULL;
    return;
  }

  if (ERR_GPIO_INVALID==(gpio=relays_get_gpio_by_idx(idx))) {
    mg_rpc_send_errorf(ri, 400, "No relay at idx=%d", idx);
    ri = NULL;
    return;
  }

  relays_set_by_idx(idx, state);
  mg_rpc_send_responsef(ri, "{idx: %d, gpio: %d, state: %d}", idx, gpio, state);

  (void) cb_arg;
  (void) fi;
}

static void rpc_relay_toggle_handler(struct mg_rpc_request_info *ri, void *cb_arg, struct mg_rpc_frame_info *fi, struct mg_str args) {
  int idx;
  uint8_t gpio;
  bool state;

  rpc_log(ri, args);

  if (json_scanf(args.p, args.len, ri->args_fmt, &idx) != 1) {
    mg_rpc_send_errorf(ri, 400, "idx is required");
    ri = NULL;
    return;
  }

  if (ERR_GPIO_INVALID==(gpio=relays_get_gpio_by_idx(idx))) {
    mg_rpc_send_errorf(ri, 400, "No relay at idx=%d", idx);
    ri = NULL;
    return;
  }

  state=!relays_get_by_idx(idx);
  relays_set_by_idx(idx, state);
  mg_rpc_send_responsef(ri, "{idx: %d, gpio: %d, state: %d}", idx, gpio, state);

  (void) cb_arg;
  (void) fi;
}

void rpc_init() {
  struct mg_rpc *c = mgos_rpc_get_global();
  mg_rpc_add_handler(c, "Button.Touch", "{idx: %d}", rpc_button_touch_handler, NULL);
  mg_rpc_add_handler(c, "Relay.Get", "{idx: %d}", rpc_relay_get_handler, NULL);
  mg_rpc_add_handler(c, "Relay.Set", "{idx: %d, state:%d}", rpc_relay_set_handler, NULL);
  mg_rpc_add_handler(c, "Relay.Toggle", "{idx: %d}", rpc_relay_toggle_handler, NULL);
}
